$large
$nofloatcalls
c  This is an experimental program to fit surfaces to data based on the
c projection pursuit approximation. The data is in the form
c
c         (Y , Z  . .. Z  )
c           1   11      1m 
c 
c 
c         (Y , Z  . .. Z   )
c           n   n1      nm
c 
c     and is read in from a file name given on the last line from the 
c     standard input ( see test input file).
c      The Z's are independent variables and Y is the dependent variable.
c     The case this program considers is when Y is a 0,1 Bernoulli R.V.
c
c                             f(z)
c     with P( Y=1| Z=z) =    e
c                         --------   .
c                              f(z)
c                         1 + e
c
c The goal is to estimate f without making any parametric assumptions 
c about the shape of this function. 
c
c The projection pursuit approximation to the function p(x) where 
c  x is an m-dimensional vector is
c                                      T 
c      f(z) = sum k from 1 to M { f ( a  z) }
c                                  k   k
c                 T
c     subject to a a = 1.
c
c
c     For a given M, both the ridge functions, f   and the directions a
c                                               k                      k
c     need to be estimated. 
c 
c
c  In this program the ridge functions are found using spline estimates
c   for generalized linear models. 
c  The projections are found by minimizing a weighted residual 
c   sum of squares.
c                                    ^  2
c    WSS = sum j from 1 to n { (Y  - Y )  / var (Y ) }
c                                j    j           j
c
c  Both estimates are pursued simultaneously using an iterative algorithm
c   known as backfitting. An outline of  this procedure is given below
c
c
c      Loop through specified number of backfits:
c               Do im=1,M
c                    Minimize the weighted residual sum of squares
c                     for the im projection. 
c               
c      End loop 
c
c     The first minimization for each direction is done using a coarse grid 
c  search and then refined using a simplex search. Subsequent searches 
c  just use the simplex refinement. One aspect of these computations 
c  that may be confusing is the implicit estimation of the ridge functions
c  themselves. Note that each candidate direction, a, yields a new dependent 
c  variable: px= Za. In order to calculate the residuals, Y is regressed on 
c  this variable subject to the constraints of the model. These constraints 
c  are not trival as they involve the logistic link function and the 
c  adjustment for the other ridge functions. 
c      
c     The smoothing parameter (bandwidth) is not estimated in this 
c  program and must be specified. Provisions have been made, however, for 
c  a search over values of the smoothing parameter but for the present I 
c  reccommend that ngl=1, xl1=xl2= fixed value for lambda. 
c                            
c***** projection pursuit glim models
c
c******* definitions of key variables
c
c       nobs - number of observations
c       x    - independent variables
c       y    - dependent variable
c       irep - the number of replications (set to one in this version)
c       a    - matrix of ridge vectors
c       ndim - number of independent Z-variables for each observation
c       ndst - number of directions in grid search 
c       mrid - number of ridge functions
c	link - link function, 1= normal 2= bernoulli 3=poisson
c
c       x is an nobs by ndim matrix
c       a is an ndim by mrid  matrix
c       snuf - current estimate of f at the observation points
c       snuc - snuf minus a particular ridge function ( this is 
c                                       continually updated)
c        
c
c****** note on input and output
c This program was developed under UNIX and reads and writes to the 
c standard devices. For the format of the input refer to the sample input
c file provided with the source listing. 
c  
c ****** A unix signal handler to find overflows, etc

c      integer function handler(sig, code, sigcontext)
c        integer sig
c        integer code
c        integer sigcontext
c        print *, 'ieee exception code', loc(code), 'occurred at pc ', 
c     *   sigcontext(4)
c      call abort
c      end

      implicit double precision (a-g,o-z)
      parameter (nmax=800,mbig=4,nd=12)
      real*8 x(nmax,nd),y(nmax),sy(nmax)
      real*8 temps(nmax),SYOLD(NMAX),MEANSY(NMAX),SESY(NMAX)
      real*8 snu(nmax),snuc(nmax),snuf(nmax),YOLD(NMAX)
      real*8 tempc(nmax),savnu(nmax,mbig)
      real*8 savpx(nmax,mbig), sout(20,mbig)
      real*8 py(nmax),px(nmax),crs(nd),svec(nd)
      real*8 a(nd,mbig),MEANA(nd,mbig),SEA(nd,mbig),OLDA(nd,mbig)
      REAL*8 OCV(mbig),GCV(mbig),LEV(nmax,mbig),GCVF,OCVF,LEVF(nmax)
      REAL*8 NUTOY,PHI2,STEP,LMIN,LMAX,YHATI,ERRVAR,
     *  tr,TRFULL,GCVTOT,OCVTOT,DIAG,LL,AVGY
      real*8 par(20),dout(20)
      integer job,irep(nmax),tempr(nmax),index(nmax),IZERO(ND,MBIG)
      integer seed, SAVSEED, csrch,NBOOT,IBOOT,NREP,ITEST

c**** set up sig handler
c       integer ieeer
c       ieeer = ieee_handler('set','common', handler)
C       ieeer = ieee_handler('set','underflow', handler)
c       if (ieeer .ne. 0) print *,'ieee_handler can not set underflow'


c**** read in 16 parameters and seed from the standard input 
c       (filename is read in subsequent call to gdata)

        call gpars(npar,par,SAVseed)
        if (npar.eq.0) stop

c***** redefine the parameter vector 

c number of Z-variables
        ndim=  int(par(1))

c number of directions in grid search
        ndst=  int(par(2))

c number of ridge functions (projections)
C test the significance of the mrid'th projection if mrid is negative
C Start by reducing the number of ridge functions by 1 to get predicted
C values for bootstrapping
	ITEST=0
	IF(PAR(3).LT.0) ITEST=1
	PAR(3)=ABS(PAR(3))
        mrid=  int(par(3))
	IF(MRID.EQ.0) MRID=1

c initial size of search region for simplex
        eps=       par(4)

c number of backfit iterations
        nback= int(par(5))
	IF(MRID.EQ.1) NBACK=2

c number of iterations in simplex search
        nit=   int(par(6))

c the next three paramters were included for possible grid searches over lambda
c (ngl=number of lambdas, x1l and x2l are lower and upper limit to range)
c CHANGED BY DS:
c I set ngl to 1 always because otherwise the subroutines go through the
c ngl lambdas with each call. Instead, I fix the subroutines to run only
c ONCE for each value of lambda. NLAM, LMIN, NMAX now define a grid
c search controlled in the main program (ie here), and no grid search is done 
c in the subroutines.
c Too arrange this alteration I have temporarily let par(7)=NLAM, 
c par(8)=LMIN and par(9)=LMAX when reading input. par(7) is then set to 1,
c meaning no grid searches in subroutines. par(8) and par(9) are later set
c to the specific (single) value of lambda redefined at each step in the 
c grid search.

CCC   ngl=   int(par(7))
CCC   x1l=       par(8)    
CCC   x2l=       par(9)

 	NLAM=  INT(PAR(7))

        NGL=       1
 	PAR(7)=    1
        LMIN=      PAR(8)
	LMAX=      PAR(9)

c these next parameters pertain to fitting a GLim spline 

        maxit=     par(10)
        stop=      par(11)
        job=       par(12)
        link=      par(13)    

c replicate measurements present? 0=no 1=yes, need extra variable at end
        lrrep= int(par(14))

c parameter for controlling the amount of output
        iptr=  int(par(15))

c same as eps but used after first coarse search
        eps2=      par(16)
	NBOOT= INT(PAR(17))
	IF(ITEST.NE.0) THEN
	  IF(NLAM.GT.1) THEN
	    WRITE(*,'(/A/A)')
     * ' ERROR: I bootstrap directions only when No. of lambdas = 1',
     *    ' Start again'
	    STOP
          ELSEIF(NBOOT.LT.2) THEN
	    WRITE(*,'(/a/a,a/a)') 
     *  ' You requested a significance test of the m''th projection.',
     *  ' You need to indicate the number of bootstrap replicates',
     *  ' desired (> 1).',
     *  ' Start again.'
	    STOP
	  ENDIF
	ENDIF
	
	IF(NLAM.GT.1.AND.NBOOT.GT.1) THEN
	    WRITE(*,'(/A/A)') 
     *    ' ERROR: I do bootstraps only when No. of lambdas = 1',
     *    ' Start again'
	    STOP
	ENDIF

c SAVSEED is the last number read by gpars

c par(18) doesn't seem to exist, so is 0 (DS).
c numcr indicates whether a coarse search should be done or not.
c Adjust value of number of coarse grid searches ( default is that
c grid search done just on the first iteration of backfitting algorithm).
        numcr= int(par(18))
        if (numcr.eq.0) numcr=1
        if (ndst.eq.0) numcr=0

c additive model flag (nit=par(6))
        iadd=0
        if (nit.eq.0) then 
              iadd=1
        endif 

c finda routine needs a higher print flag to print intermediate results
        iptrf= iptr -1 


c if ndst equals zero read in starting value for the directions.
        if( ndst.eq.0) then 
          do 2 im=1,mrid
            read(*,*) (a(j,im),j=1,ndim) 
    2     continue
          write(*,*) '#  estimates in the first iteration',
     *     'being computed using starting values'
        endif


        ldx= nmax

c read in data: data file name is read from the standard input 
c lrrep >0 indicates that replicate numbers are included in the file
c as the last column of each row

        call gdata(ldx,nobs,ndim,x,lrrep, irep,y,LINK,NREP)
	do 4 k=1,nobs
	  YOLD(k)=Y(k)
4	continue

c write out the parameters

        WRITE(*,*) '# Parameters read in: '
	WRITE(*,201) '   No. of Z-variables =                  ',ndim
	WRITE(*,201) '   No. of directions in grid search =    ',ndst
	WRITE(*,201) '   No. of projections                    ',mrid
	WRITE(*,201) '   Link (1=norm, 2=binom, 3=poiss) =     ',link
c	WRITE(*,200) '   eps =                                 ',eps
c	WRITE(*,200) '   eps2 =                                ',eps2
	WRITE(*,201) '   No. of backfit iterations =           ',nback
	WRITE(*,201) '   No. of iterations in simplex search = ',nit
	WRITE(*,201) '   Max no. of iterations for spline fit =',maxit
	IF (NLAM.EQ.1) then
	 IF(NBOOT.GT.1) WRITE(*,201)
     *                '   No. of bootstrap replicates =         ',NBOOT
	 WRITE(*,200) '   Lambda =                              ',LMIN
	ELSE
	 WRITE(*,201) '   No. of lambdas =                      ',NLAM
	 WRITE(*,200) '   Lower lambda =                        ',LMIN
	 WRITE(*,200) '   Upper lambda =                        ',LMAX
	 IF(LMIN.GE.LMAX) THEN
	   WRITE(*,'(/a/a/)') 
     *     ' ERROR: Upper lambda not larger than lower lambda',
     *     ' No. of lambdas set to 1'
	   NLAM=1
	   LMAX=LMIN
	 ENDIF
	ENDIF
	WRITE(*,201) '   Seed for random projection search =   ',SAVseed

200	FORMAT(a,6f10.6)
201     FORMAT(a,i10)

	IF (NLAM.GT.1) THEN
          STEP=(LMAX-LMIN)/(NLAM-1)
	ELSE
	  STEP=0.0
	ENDIF

c Begin the loop over values of lambda (x1l, x2l, par(8) and par(9)
c are redefined here).

c NB: seed is the starting integer used by the random number generator.
c For different lambda values, seed is reset to SAVSEED at the start
c of each grid search (to minimize the role of seed when assessing the
c benefits of each value of lambda). The whole process should be run
c several times for different seeds.

4999  CONTINUE

      DO 5000 L=1,NLAM
      seed=SAVSEED

c   Start the bootstrapping loop
       DO 4000 IBOOT=0,NBOOT

	x1l=(L-1)*STEP + LMIN
	x2l=x1l
	par(8)=x1l
	par(9)=x1l

       IF(IBOOT.EQ.0) THEN
	 IF(ITEST.EQ.1) THEN
	   WRITE(*,'(/a/a,i2,a,a/a,i2,a)')
     * '# ===========================================================',
     * '  Significance test of projection',mrid,' starts with an',
     * '  analysis','  of the data using',mrid-1,' projection(s):'
	   MRID=MRID-1
	   PAR(3)=MRID
	   IF(MRID.EQ.0) THEN
	     WRITE(*,*) 
     *    '(Note: for 0 projections we use average Y to calculate Yhat)'
	     AVGY=0.0
	     DO 5 K=1,NOBS
	       AVGY=AVGY+Y(K)*DBLE(IREP(K))
5	     CONTINUE
	     AVGY=AVGY/DBLE(NREP)
cbug
 	       write(*,*) 'Average Y =',avgy
	     DO 6 K=1,NOBS
	       SYOLD(K)=AVGY
	       TEMPS(K)=Y(K)-AVGY
6	     CONTINUE
	     IF(LINK.EQ.1) ERRVAR=VARIANC(TEMPS,IREP,NOBS)
	     CALL LOGLIK(Y,SYOLD,IREP,NOBS,LINK,ERRVAR,LL)
	     WRITE(*,'(/A,F13.5)') 'Log likelihood = ',LL
	     GOTO 100
	   ENDIF
	 ENDIF

	 IF(NLAM.GT.1) WRITE(*,'(a/a/a,f10.6/)')
     * '# ===========================================================',
     * ' ',
     * '# Results for lambda =   ',x1l

	 IF(LINK.NE.1) THEN 
	   WRITE(*,'(/a)')
     *    '    Backfit  Projection    Weighted SSR  #iterations'
	 ELSE
	   WRITE(*,'(/a)')
     *    '    Backfit  Projection    Weighted SSR'
	 ENDIF

       ELSEIF(IBOOT.EQ.1.AND.ITEST.NE.0) THEN
	   MRID=MRID+1
	   PAR(3)=MRID
	   WRITE(*,'(/a/a/a,i2,a/a)')
     * '# ===========================================================',
     * '  Next, bootstrap samples are created from the above estimate',
     * '  and analysed using',mrid,' projection(s). Listed below are:',
     * '  Bootstrap replicate, No.parameters, GCV score, Loglikelihood'

       ELSEIF(IBOOT.EQ.1.AND.ITEST.EQ.0) THEN
	  WRITE(*,'(/a/a/a)')
     * '# ===========================================================',
     * '  Creating bootstrap samples from the above estimate:',
     * '  Bootstrap replicate, projection no., projection coefficients'

       ENDIF

c  Initialize the vectors that will contain the estimates of f(Z) and 
c  the individual ridge functions for each observation 
c  snuf = current estimate of f at the observation points
c  snuc = snuf minus a particular ridge function (this is continually updated)
c 
        do 10 k=1,nobs
          snuc(k)=0.0
          snu(k)=0.0
          snuf(k)=0.0
	  LEVF(k)=0.0
	  DO 11 im=1,mrid
	    SAVNU(k,im)=0.0
	    LEV(k,im)=0.0
   11     CONTINUE
   10   continue


c  begin the loop for backfitting 
	
       do 2000 iback= 1,nback

c  decide if there will be a coarse search on the ridge vector
c  (is yes only at the first backfit, iback=1, and not even then in special
c  circumstances, such as when ndst=0)

       if( (iback.le.numcr).and.(ndst.ne.0)) then
            csrch=1
       else
            csrch=0
       endif

c  begin loop through the ridge vectors 

         do 1000 im=1,mrid

c  Update the snuc functions and copy starting direction for call to finda 
            do 25 k=1,nobs
               snuc(k)= snuf(k)- savnu(k,im)
   25        continue
            do 27 j=1,ndim
               crs(j)= a(j,im)
   27       continue

c  call finda if no start values are given (ndst>0 ) or
c  if this is not the first iteration of the loop
c  or if the additive model flag is true
c  (iadd=1 if the number of simplex searches is zero)

       if( (ndst.ne.0).or.(iback.gt.1).or.(iadd.ne.1) ) then

c         write(*,*) '>  call to finda on backfit ',iback,
c    *     'for ridge function ',im
c         do 9000 k=1,nobs
c             write(*,9005) k,snuf(k), snuc(k)
c9005         format( 5e15.5)
c9000      continue

          call finda(ldx,nobs,ndim,x,y,irep,snuc,par,
     *      crs,svec,dout,seed,csrch,iptrf,LEV(1,im))
        
       else 

c       if starting directions have been supplied 
c       use them to estimate the ridge function 
 
           dout(2)= exp(x1l)
           do 28 k =1,ndim
               svec(k)= a(k,im)
  28       continue
       endif

c   write out intermediate results if iptr > 0

       if(iptr.ge.1) then
          write(*,*) '> best projection  number',im 
          write(*,300) (svec(k),k=1,ndim)
          write(*,300) ( dout(k),k=1,4)
 300      format('>',6e12.4)
       endif

c   Find the estimate for f(Z) and the predicted values at this best direction
c   This step is a lazy one because finda has already done this computation 
c   but the results were not passed back. It would have complicated the 
c   calling arguments to finda and subsequent subroutines.

c   compute px= Za  and put these values along with y, snuc, and irep
c       in sorted order

c       note: since mkpx sorts on px this is the reason for
c       temporary variables and the variable index ( indexes sorted order) 
c

            call mkpx(ndim,svec,ldx,nobs,x,y,snuc,irep,
     *           px,py,tempc,tempr,index)

c compute glim spline best value.
C DS has added LEV to the list of arrays passed back from glisp, so that
C GCV and OCV scores can be calculated subsequently. Note that LEV(1,im)
C is the leverages from a glisp smooth of py-tempc on px, NOT of py on px
C (tempc = snuc in glisp). That is, LEV is a regression of residuals of py
C (after subtracting the yhats from all other predictor x-variables) on px.

c itr controls whether trace is computed in smths (1=yes, 0=no)

            itr=1

            call glisp(dout(2),nobs,px,py,temps,
     *      snu,tempc,fitss,tr,WSS,
     *      maxit,stop,numit,tstop,0,link,itr,tempr,LEV(1,im))

c   write out some useful parameters of this fit
c   WSS= weighted RSS, numit = #iterations for convergence of GL spline
c   procedure ( compare with maxit), tstop= convergence criterion 
c   (compare with stop), tr= trace of A matrix= effective number of
c   parameters used in this fit to the ridge function
c   fitss = fit relative to a constant

             if(iptr.GE.1) then
	       WRITE(*,'(a)')
     *    '    Backfit  Projection    Weighted SSR  #iterations  trace'
	       WRITE(*,'(i8,i10,f21.8,i7,f10.5)') iback,im,WSS,numit,tr
	     ELSEIF(IBOOT.EQ.0) THEN
		IF(LINK.NE.1) THEN
	         WRITE(*,'(i8,i10,f21.8,i7)') iback,im,WSS,numit
		ELSE
	         WRITE(*,'(i8,i10,f21.8)') iback,im,WSS
		ENDIF
             endif

c   update projection vector  

            do 15 j=1,ndim
               a(j,im)=svec(j)
   15       continue
           
c save various values 

            sout(1,im)= WSS
            sout(2,im)= dout(2)
            sout(3,im)= tr
            sout(4,im)= numit
            sout(5,im)= tstop
            sout(6,im)= fitss

c   save estimated ridge function and update the estimate of f(Z)
c   indexing is necessary because the array has been scrambled due to 
c   sorting on px above
c   Note that temps is the full yhat  (and sy will be too after stmt below)

            do 20 k=1,nobs
               kk= index(k)
               savnu(kk,im)= snu(k)
               savpx(kk,im)= px(k)
               sy(kk)= temps(k)
               snuf(kk)= snuc(kk) + snu(k)
   20       continue


cbug         do 21 k=1,nobs
cbug         write(*,3005) (savnu(k,jj), jj=1,mrid), snuf(k),y(k)
cbug   21    continue

1000      continue
2000    continue


c Compute the log likelihood (compute ERRVAR for normal data if IBOOT=0)
	IF(LINK.EQ.1.AND.IBOOT.EQ.0) THEN
	  do 30, k=1,nobs
	    TEMPS(K)=SY(K)-Y(K)
30	  continue
	  ERRVAR=VARIANC(TEMPS,IREP,NOBS)
	ENDIF
	CALL LOGLIK(Y,SY,IREP,NOBS,LINK,ERRVAR,LL)


c Compute the OCV and GCV scores. 
C NB: Note that OCV score as computed below will be incorrect
C     if adjacent data points had to be combined in smths because
C     of ties or near-ties. This is because LEV returned from smths
C     has fewer cases than the full y vector. Delete the OCV score?
C ** this not a problem anymore since I no longer combine ties in smths,
C    see modification to glisp to spread apart near-ties.

c The following computes GCV and OCV scores for each projection
c separately: how well does savnu(im) predict y?

       IF(IBOOT.ge.0) THEN
C{begin gcv and ocv}
        DO 46 im=1,MRID
	 OCV(im)=0.0
	 GCV(im)=0.0

         DO 45 k=1,nobs

	    IF(LINK.EQ.1) THEN
	      YHATI=SAVNU(k,im)
	      DIAG=1.0D0
	    ELSE
              YHATI=NUTOY(SAVNU(k,im),link)
              DIAG=PHI2(SAVNU(k,im),link)
	    ENDIF

            OCV(im)=OCV(im) + 
     *       (DBLE(IREP(k))/DIAG)*( (Y(k)-YHATI)/(1-LEV(k,im)) )**2.
            GCV(im)=GCV(im) + 
     *       ( DBLE(IREP(k))/DIAG ) *  ( Y(k)-YHATI )**2.

45       CONTINUE

         GCV(im)=DBLE(NOBS)*GCV(im)/(DBLE(NOBS)-SOUT(3,im))**2.
         OCV(im)=OCV(im)/DBLE(NOBS)

46      CONTINUE
 
c Now compute the full GCV and OCV scores

        GCVTOT=0
        OCVTOT=0
        GCVF=0
        OCVF=0
        TRFULL=0.0

        DO 50 im=1,MRID
	 TRFULL=TRFULL+sout(3,im)
	 GCVTOT=GCVTOT+GCV(im)
	 OCVTOT=OCVTOT+OCV(im)
50      CONTINUE

        DO 60 k=1,nobs
	  IF(LINK.EQ.1) THEN
	    YHATI=SNUF(k)
	    DIAG=1
	  ELSE
            YHATI=NUTOY(SNUF(k),link)
            DIAG=PHI2(SNUF(k),link)
	  ENDIF

	  DO 61 im=1,MRID
	    LEVF(k)=LEVF(k)+LEV(k,im)
61	  CONTINUE

          OCVF=OCVF + 
     *       (DBLE(IREP(k))/DIAG)*( (Y(k)-YHATI)/(1-LEVF(k)) )**2.
          GCVF=GCVF + 
     *       ( DBLE(IREP(k))/DIAG ) *  ( Y(k)-YHATI )**2.

60       CONTINUE

         GCVF=DBLE(NOBS)*GCVF/((DBLE(NOBS)-TRFULL)**2.0)
         OCVF=OCVF/DBLE(NOBS)


c note that GCVF is just: dble(nobs)*wss/((dble(nobs)-trfull)**2.0)
c ie there's no need to calculate the stuff in loop 60 for GCVF as it was
c already calculated in glisp (this isn't true of OCVF) 
	 IF(IBOOT.EQ.0) THEN
           WRITE(*,*) 
           IF(MRID.GE.3) THEN
	     WRITE(*,*)
     * '    Variable Projection 1    Projection 2    Projection 3  ...'
           ELSEIF(MRID.EQ.2) THEN
	    WRITE(*,*)
     * '    Variable Projection 1    Projection 2'
           ELSE
	    WRITE(*,*)
     * '    Variable Projection 1'
           ENDIF
           do 2900 k=1,ndim
	   write(*,'(i8,1x,f13.5,8(3x,f13.5))')  k,(a(k,im), im=1,mrid)
2900       continue

	   IF(MRID.GT.1) WRITE(*,'(/a,1x,f10.5,8(3x,f13.5))')
     * '    GCV(i) ',(GCV(im),im=1,mrid)

           WRITE(*,*) ' '
           WRITE(*,*) 
     * '    No. parameters   GCV score   Log likelihood'
           WRITE(*,'(f13.5,3(3x,f13.5))') TRFULL,GCVF,LL
           WRITE(*,*) ' '
	 ELSEIF(ITEST.EQ.1) THEN
           WRITE(*,'(i5,f13.5,3(3x,f13.5))') IBOOT,TRFULL,GCVF,LL
	 ENDIF
C{end gcv and ocv}
       ENDIF

        IF(NLAM.EQ.1.AND.IBOOT.EQ.0.AND.ITEST.EQ.0) THEN
	  WRITE(*,*) ' '
	  IF(MRID.EQ.1) THEN
	   WRITE(*,'(7x,a)') 
     *  'Y          Yhat         pX1         f1'
	  ELSEIF(MRID.EQ.2) THEN
	   WRITE(*,'(7x,a,a)') 
     *  'Y          Yhat         pX1         f1',
     *  '          pX2         F2'
	  ELSE
	   WRITE(*,'(7x,a,a)') 
     *  'Y          Yhat         pX1         f1',
     *  '          pX2         F2          pX3         F3 ...'
	  ENDIF
	  do 3000 k=1,nobs
	   WRITE(*,'(8f12.5)') 
     *     y(k),sy(k),(savpx(k,im),savnu(k,im),im=1,mrid)
 3000     continue
        ENDIF

	IF(NLAM.EQ.1) THEN

	  IF(IBOOT.EQ.0) THEN
c         Set arrays to zero on first pass through

	    DO 85 K=1,NOBS
	      SYOLD(K)=SY(K)
	      MEANSY(K)=0.0
	      SESY(K)=0.0
85	    CONTINUE
	    DO 86 K=1,NDIM
	      DO 87 IM=1,MRID
		OLDA(K,IM)=A(K,IM)
		MEANA(K,IM)=0.0
		SEA(K,IM)=0.0
87	      CONTINUE
86	    CONTINUE

	  ELSEIF(ITEST.EQ.0) THEN
c         Point randomly generated "a" in the right direction

	    DO 90 IM=1,MRID
	      IF( ACORR(OLDA(1,IM),A(1,IM),NDIM,0) .LT.
     *           ACORR(OLDA(1,IM),A(1,IM),NDIM,1) ) THEN
		DO 88 K=1,NDIM
		  A(K,IM)=-A(K,IM)
88		CONTINUE
	      ENDIF

C             Check whether a(ij) is of different sign than olda(ij)
	      DO 89 K=1,NDIM
		IF(OLDA(K,IM)*A(K,IM).LE.0) 
     *               IZERO(K,IM)=IZERO(K,IM)+1
89	      CONTINUE

90	    CONTINUE

	    DO 91 IM=1,MRID
	     WRITE(*,'(2I5,12f9.5))') 
     *         IBOOT,im,(a(k,im), K=1,NDIM)
91	    CONTINUE
	    
C   Accumulate sums of yhat and yhat**2 (nu scale), and of "a"
	    IF(LINK.GT.1) THEN
	      DO 92 K=1,NOBS
		MEANSY(K)=MEANSY(K)+YTONU(SY(K),LINK)
		SESY(K)=SESY(K)+(YTONU(SY(K),LINK))**2
92	      CONTINUE
	    ELSE
	      DO 93 K=1,NOBS
		MEANSY(K)=MEANSY(K)+SY(K)
		SESY(K)=SESY(K)+SY(K)**2
93	      CONTINUE
	    ENDIF

	    DO 97 IM=1,MRID
	      DO 96 K=1,NDIM
cbug      write(*,*) k,im,a(k,im)
		MEANA(K,IM)=MEANA(K,IM)+A(K,IM)
		SEA(K,IM)=SEA(K,IM)+A(K,IM)**2
96	      CONTINUE
97	    CONTINUE
	  ENDIF
	ENDIF

100	IF(NLAM.EQ.1.AND.IBOOT.LT.NBOOT) THEN
	    CALL REBOOT(Y,SYOLD,IREP,NOBS,LINK,ERRVAR,SEED)
	ENDIF

C      {end of bootstrap loop}
4000   CONTINUE

5000  CONTINUE

C      Loop back one more time to fit the mrid functions and
C      and give the output, including the log-likelihood
C      note that SAVSEED will be reused. Note also that since
C      Y has been changed by the previous bootstraps, it needs to
C      be reset to its original values
      IF(ITEST.EQ.1) THEN
	ITEST=0
	NBOOT=0
	WRITE(*,'(/a/a,i2,a,a/a/a,i2,a,a/a)')
     * '# ===========================================================',
     * '  Finally, the real data are fit with',mrid,' projection(s).',
     * ' If the ',  
     * '  loglikelihood below is greater than 95% of the bootstrap',
     * '  values above, then projection',mrid,' is deemed',
     * ' statistically',
     * '  significant at P=0.05'
	DO 101 K=1,NOBS
	  Y(K)=YOLD(K)
101	CONTINUE
	GOTO 4999
      ENDIF

      IF(NBOOT.GT.1.AND.ITEST.EQ.0) THEN
	WRITE(*,'(//A/A)')
     * '# ===========================================================',
     * '  Summary of bootstrap results:'

	DO 102 K=1,NDIM
	  DO 103 IM=1,MRID
	    SEA(K,IM)=SQRT( (SEA(K,IM)-(MEANA(K,IM)**2)/NBOOT)/NBOOT )
	    MEANA(K,IM)=MEANA(K,IM)/NBOOT
103	  CONTINUE
102	CONTINUE

        WRITE(*,'(/a/)') '# Mean of bootstrap replicate coefficients'
        IF(MRID.GE.3) THEN
	  WRITE(*,*)
     * '    Variable Projection 1    Projection 2    Projection 3  ...'
        ELSEIF(MRID.EQ.2) THEN
	  WRITE(*,*)
     * '    Variable Projection 1    Projection 2'
        ELSE
	  WRITE(*,*)
     * '    Variable Projection 1'
        ENDIF
         do 120 k=1,ndim
	  write(*,'(i8,1x,f13.5,8(3x,f13.5))')  
     *      k,(meana(k,im), im=1,mrid)
120     continue
        WRITE(*,'(/a/)') '# SE of bootstrap replicate coefficients'
        IF(MRID.GE.3) THEN
	  WRITE(*,*)
     * '    Variable Projection 1    Projection 2    Projection 3  ...'
        ELSEIF(MRID.EQ.2) THEN
	  WRITE(*,*)
     * '    Variable Projection 1    Projection 2'
        ELSE
	  WRITE(*,*)
     * '    Variable Projection 1'
        ENDIF
         do 130 k=1,ndim
	  write(*,'(i8,1x,f13.5,8(3x,f13.5))')  
     *      k,(sea(k,im), im=1,mrid)
130     continue

        WRITE(*,'(/a/a/)') 
     *   '# No. of times bootstrap coefficient was of different sign',
     *   '  than the original estimate'
        IF(MRID.GE.3) THEN
	  WRITE(*,*)
     * '    Variable Projection 1    Projection 2    Projection 3  ...'
        ELSEIF(MRID.EQ.2) THEN
	  WRITE(*,*)
     * '    Variable Projection 1    Projection 2'
        ELSE
	  WRITE(*,*)
     * '    Variable Projection 1'
        ENDIF
         do 140 k=1,ndim
	  write(*,'(i8,1x,i13,8(3x,i13))')  
     *      k,(izero(k,im), im=1,mrid)
140     continue

	WRITE(*,'(/6X,A)')
     *    'Yhat      Yhat-1se    Yhat+1se  Bootstrap mean' 
	IF(LINK.EQ.1) THEN
	 DO 150 K=1,NOBS
	   SESY(K)=SQRT( ( SESY(K)-(MEANSY(K)**2)/NBOOT )/NBOOT )
	   MEANSY(K)=MEANSY(K)/NBOOT
	   X1= SYOLD(K)-SESY(K)
	   X2= SYOLD(K)+SESY(K)
	   X3= MEANSY(K)
	   WRITE(*,'(4F12.5)') SYOLD(K),X1,X2,X3
150	 CONTINUE
	ELSE
	 DO 160 K=1,NOBS
	   SESY(K)=SQRT( ( SESY(K)-(MEANSY(K)**2)/NBOOT )/NBOOT )
	   MEANSY(K)=MEANSY(K)/NBOOT
	   X1= YTONU(SYOLD(K),LINK) - SESY(K)
	   X1= NUTOY(X1,LINK)
	   X2= YTONU(SYOLD(K),LINK) + SESY(K)
	   X2= NUTOY(X2,LINK)
	   X3= NUTOY(MEANSY(K),LINK)
	   WRITE(*,'(4F12.5)') SYOLD(K),X1,X2,X3
160	 CONTINUE
	ENDIF

      ENDIF
      
      end



c This is the main optimization routine called by the main program
c The structure is to minimize the objective function using a coarse
c search over points randomly chosen on an ndim sphere. The search is
c then refined using a simplex search. The simplex search has been
c modifed to constrain the direction vector to always have unit length. 
c As might be expected there are several tuning parameters to this 
c algorithm. The most important being the number of points considered 
c in the coarse search (ndst) and the number of iterations of the 
c simplex search (nit).
c 
c***** projection pursuit glim models
c
      subroutine finda(ldx,nobs,ndim,x,y,irep,snuc,par,
     *      crs, svec,dout,seed,ic,iptr,LEV)
      implicit double precision (a-g,o-z)
      parameter (nmax=800,nd=12,nproj=10,ntest=10)
      real*8 x(ldx,ndim),y(nobs),snuc(nobs),par(1),LEV(NMAX)
      real*8 direct(nd) ,py(nmax),px(nmax),tempc(nmax),svec(ndim)
      real*8 crs(ndim)
      real*8 objfun(nproj),vertex(nproj,nd),test(ntest,nd)      
      real*8 ht(nproj),hl(nproj)      
      real*8 tout(20),dout(1)
      integer job,seed,irep(nmax),tempr(nmax),index(nmax)
      integer worst, best
     
        ldir= nproj

        ndim= int(par(1))
        ndst= int(par(2))
        mrid= int(par(3))
        eps= par(4)
        nback= par(5)
        nit= int(par(6))
        ngl= int(par(7))
        x1l= par(8)    
        x2l= par(9)
        maxit= par(10)
        stop= par(11)
        job= par(12)
        link= par(13)    
        eps2= par(16)
  

c NB: iptr set to iptr-1 in main routine

      if( ic.eq.0) then 

c******* find objective function at start values
c****** then optimize over lambda for the fixed direction.

          call mkpx(ndim,crs,ldx,nobs,x,y,snuc,irep,
     *       px,py,tempc,tempr,index)

          call  gcvpx(nobs,tempr,px,py,tempc,par(7),tout,LEV)
          gcrs= tout(1)
          if(iptr.ge.1) then 
          write(*,*) '# starting direction' 
          write(*,305) gcrs,(crs(k) ,k=1,ndim)
          endif
      else 

c********* coarse search to find best best direction
        
        job=0
        if( iptr.ge.1) then 
         write(*,*) 
     *    ' coarse search in finda; direction#, WSS, direction'
        endif 
           call ransh(ndim,direct,seed)
      
        do 1000 idir=1,ndst
c******     make up projection vector
            call ransh(ndim,direct,seed)
            call mkpx(ndim,direct,ldx,nobs,x,y,snuc,irep,
     *        px,py,tempc,tempr,index)
c******     optimize over lambda for each fixed direction.
cbug           write(*,*) ' call to gcvpx in finda'
            call  gcvpx(nobs,tempr,px,py,tempc,par(7),tout,LEV)
            WSS= tout(1)
            tr= tout(3)
          
            if(iptr.ge.1) then
              write(*,300)  idir, WSS
              write(*,301)  (direct(k) ,k=1,ndim)
            endif

            if( (idir.eq.1).or.( WSS.lt.gtest))  then
              gtest=WSS
              itest= idir
              ttest= tr
              do 400 j=1,ndim
                crs(j)= direct(j)
  400         continue
            endif
 1000   continue

        gcrs= gtest

      endif 


        if((iptr.ge.1).and.(ic.eq.1) ) then 
         write(*,*) '# best point on grid search =', itest, 
     *      ', best WSS =',gcrs
         write(*,305) (crs(k) ,k=1,ndim)
  305    format('# ',5e12.4)
  301    format('# ',6e12.4)
  300    format('# ',i4,5e12.4,/,5e15.8)
        endif

        if( ic.eq.1) then
           bsize= eps
        else
           bsize= eps2
        endif

        if(iptr.ge.2) then
          write(*,*) '# simplex search with start radius= ', bsize
        endif

         call  mkvert(ndim,crs,nproj,nv,vertex,bsize,seed)

         do 2900 k=1,nv
            call mkpxl(k,nproj,ndim,vertex,ldx,nobs,x,y,snuc,irep,
     *       px,py,tempc,tempr,index)
        
            call  gcvpx(nobs,tempr,px,py,tempc,par(7),tout,LEV)

            objfun(k)= tout(1)
            hl(k)=tout(2)
            ht(k)=tout(3)

            if( iptr.ge.2) then
            write(*,*)
     *        '#', k,objfun(k),(vertex(k,j) ,j=1,ndim)
            endif

 2900   continue

c**** use the worst point from the coarse search as a member of the
c**** simplex
         call  smplx(nproj,nv,ndim,vertex,objfun,
     *         worst,best,ntest,nt,test,size)

         if( objfun(worst).gt.gcrs) then
               do 2905 j=1,ndim
                   vertex(worst,j)= crs(j)
 2905          continue
               objfun(worst)=gcrs
         endif

c**** start iterative loop of simplex search

         do 2000 it=1,nit

            call  smplx(nproj,nv,ndim,vertex,objfun,
     *         worst,best,ntest,nt,test,size)

            if( iptr.ge.3) then 
            write(*,*) '## current set of points on simplex:'
            do 3000 k=1,nv
               write(*,300) k,objfun(k),(vertex(k,j),j=1,ndim)
 3000       continue
            endif

            if( iptr.ge.2) then
            write(*,*) '#  current best itt, objfun, size:'
            write(*,*)  '#',   it,objfun(best),size
            write(*,*) '# coordinates of current best'
            write(*,301)  (vertex(best,j) ,j=1,ndim)
            endif

            do 2010 ind=1,nt
c****** make up projection vector
           call mkpxl(ind,ntest,ndim,test,ldx,nobs,x,y,snuc,irep,
     *       px,py,tempc,tempr,index)
c****** estimate gcv spline
             call  gcvpx(nobs,tempr,px,py,tempc,par(7),tout,LEV)
             if( iptr.ge.3) then
             write(*,*) '##  test line search:', ind, tout(1),tout(2)
             write(*,301) (test(ind,j) ,j=1,ndim)
             endif

              if( (ind.eq.1).or.( tout(1).lt.gtest))  then
                gtest=tout(1)
                testl=tout(2)
                testt=tout(3)
                itest= ind
               endif

 2010      continue

           objfun(worst)=gtest
           ht(worst)= testt
           hl(worst)=testl

           do 2020 j=1,ndim
              vertex(worst,j)= test(itest,j)
 2020      continue

 2000    continue

            call  smplx(nproj,nv,ndim,vertex,objfun,
     *         worst,best,ntest,nt,test,size)
            if(iptr.ge.2) then 
            write(*,*) '# ************ end of simplex search *******'
            write(*,*) '#  final direction after ', nit, ' iterations'
            write(*,301) (vertex(best,j) ,j=1,ndim)
            endif

         do 3500 j=1,ndim
          svec(j)= vertex(best,j)
 3500    continue

         dout(1)= objfun(best)
         dout(2)= hl(best)
         dout(3)= ht(best)
         dout(4) = size

      return

999   end


